mod data;
mod generators;
mod templates;

use std::fs::File;
use std::io::Read;
use clap::Parser;
use syn::spanned::Spanned;

#[derive(Debug, clap::Parser)]
struct Args {
    pub file: std::path::PathBuf,
    #[arg(short='n', long="name", help="base name for output files", required=true)]
    pub rpc_name: std::path::PathBuf,
    #[command(flatten)]
    pub generators: GenArgs
}

#[derive(Debug, clap::Args)]
#[group(required = true, multiple = true)]
struct GenArgs {
    #[arg(value_enum, short='c', long="client")]
    pub clients: Vec<generators::ClientGenerators>,
    #[arg(value_enum, short='s', long="server")]
    pub servers: Vec<generators::ServerGenerators>
}

static SOURCE_FILE: once_cell::sync::OnceCell<String> = once_cell::sync::OnceCell::new();
static SOURCE: once_cell::sync::OnceCell<String> = once_cell::sync::OnceCell::new();

fn parse_enum(item: &syn::ItemEnum) -> data::EnumTy {
    data::EnumTy {
        name: item.ident.to_string(),
        values: item.variants.iter().enumerate().map(|(i, v) | (v.ident.to_string(), i)).collect()
    }
}

fn parse_type_string(ty: String) -> data::Types {
    use data::Types::*;
    match ty.as_str() {
        "str" | "String" => String,
        "bool" => Bool,
        "f32" => F32,
        "f64" => F64,
        "i8" => I8,
        "i16" => I16,
        "i32" => I32,
        "i64" => I64,
        "u8" => U8,
        "u16" => U16,
        "u32" => U32,
        "u64" => U64,
        _ => Named(ty)
    }
}

fn parse_type(item: &syn::Type) -> data::Types {
    match item {
        syn::Type::Path(path) => {
            let segments = &path.path.segments;
            if segments.len() != 1 {
                emit_error(item.span(), "Path segments with len != 1");
            }
            let segment = &segments[0];
            if !segment.arguments.is_empty() {
                if segment.ident.to_string() != "Option" {
                    emit_error(item.span(), "Only Option are currently allowed to have arguments");
                }
                let args = match &segment.arguments {
                    syn::PathArguments::AngleBracketed(v) => v,
                    _ => emit_error(item.span(), "Angle bracketed arguments expected")
                };
                if args.args.len() != 1 {
                    emit_error(item.span(), "Expected 1 argument");
                }
                let ty = match &args.args[0] {
                    syn::GenericArgument::Type(v) => parse_type(v),
                    _ => emit_error(item.span(), "Type bracketed arguments expected")
                };
                data::Types::Optional(ty.into())
            } else {
                parse_type_string(segment.ident.to_string())
            }
        }
        syn::Type::Slice(slice) => {
            data::Types::Array(parse_type(&slice.elem).into())
        }
        _ => emit_error(item.span(), "Unsupported type")
    }
}

fn parse_struct(item: &syn::ItemStruct) -> data::StructTy {
    let name = item.ident.to_string();
    let mut fields = vec![];
    for field in &item.fields {
        if field.ident.is_none() {
            emit_error(field.span(), "Missing field name");
        }
        let name = field.ident.as_ref().unwrap().to_string();
        let ty = parse_type(&field.ty);
        fields.push(data::FieldTy { name, ty });
    }
    data::StructTy { name, fields }
}

fn try_parse_iterator(ty: &syn::Type) -> Option<data::Types> {
    if let syn::Type::Path(ty) = ty {
        let seg = ty.path.segments.last()?;
        if seg.ident.to_string() == "Iterator" {
            if let syn::PathArguments::AngleBracketed(args) = &seg.arguments {
                if let Some(syn::GenericArgument::Type(ty)) = args.args.first() {
                    Some(parse_type(ty))
                } else { None }
            } else { None }
        } else { None }
    } else { None }
}

fn parse_method(item: &syn::Signature) -> data::MethodTy {
    let mut method = data::MethodTy::default();
    method.name = item.ident.to_string();

    for arg in &item.inputs {
        let arg = match arg {
            syn::FnArg::Typed(v) => v,
            _ => emit_error(arg.span(), "Unsupported argument")
        };
        let ty = parse_type(&arg.ty);
        let name = match &*arg.pat {
            syn::Pat::Ident(v) => v.ident.to_string(),
            _ => emit_error(arg.span(), "Unsupported argument")
        };
        method.args.push(data::FieldTy { name, ty });
    }

    match &item.output {
        syn::ReturnType::Default => {
            method.ret = None;
            method.ret_stream = false;
        }
        syn::ReturnType::Type(_, ty) => {
            if let Some(ty) = try_parse_iterator(ty) {
                method.ret_stream = true;
                method.ret = Some(ty);
            } else {
                method.ret_stream = false;
                method.ret = Some(parse_type(ty));
            }
        }
    };

    method
}

fn parse_service(item: &syn::ItemTrait) -> data::ServiceTy {
    let name = item.ident.to_string();
    let mut methods = vec![];
    for item in &item.items {
        let item = match item {
            syn::TraitItem::Fn(v) => v,
            _ => emit_error(item.span(), "Only functions are supported")
        };
        methods.push(parse_method(&item.sig));
    }
    data::ServiceTy { name, methods }
}

fn main() {
    let args = Args::parse();

    let mut file = File::open(args.file.clone()).unwrap();
    let mut content = String::new();
    file.read_to_string(&mut content).unwrap();

    SOURCE_FILE.set(args.file.to_string_lossy().to_string()).unwrap();
    SOURCE.set(content).unwrap();

    let ast = syn::parse_file(SOURCE.get().unwrap()).unwrap();

    let mut rpc = data::RPC::default();

    for item in &ast.items {
        match item {
            syn::Item::Enum(v) => rpc.enums.push(parse_enum(v)),
            syn::Item::Struct(v) => rpc.structs.push(parse_struct(v)),
            syn::Item::Trait(v) => rpc.services.push(parse_service(v)),
            syn::Item::Use(_) => {}
            _ => emit_error(item.span(), "Unsupported item")
        }
    }

    for gen in &args.generators.clients {
        gen.generate(&args.rpc_name, &rpc);
    }

    for gen in &args.generators.servers {
        gen.generate(&args.rpc_name, &rpc);
    }
}

fn emit_error(span: proc_macro2::Span, msg: impl Into<String>) -> ! {
    use codespan_reporting::{
        diagnostic::{Diagnostic, Label},
        files::{SimpleFiles, Files},
        term::{
            self,
            termcolor::{ColorChoice, StandardStream},
        },
    };
    let mut files = SimpleFiles::new();
    let config = term::Config::default();
    let writer = StandardStream::stderr(ColorChoice::Auto);
    let file_id = files.add(SOURCE_FILE.get().unwrap().clone(), SOURCE.get().unwrap().clone());

    let file = files.get(file_id).unwrap();

    let start = span.start();
    let start: usize = file.line_range((), start.line-1).unwrap().start + start.column;
    let end = span.end();
    let end: usize = file.line_range((), end.line-1).unwrap().start + end.column;

    let diagnostic = Diagnostic::error()
        .with_labels(vec![Label::primary(file_id, start..end).with_message(msg)]);

    term::emit(&mut writer.lock(), &config, &files, &diagnostic).expect("cannot write error");
    std::process::abort();
}
